#!/usr/bin/env python

import numpy as np
import kepler as kp
import environment
import datetime
from mayavi import mlab
import random


def plot_n_orbits(orbits,environment,show_plot=True,save_plot=False,title='No Title'
				 ,show_body=True):
	'''
	creates a 3D plot from the input rs list, containing x, y, and z positions
	'''
	mlab.figure(1, bgcolor=(0, 0, 0), fgcolor=(0.8, 0.8, 0.8),size=(800, 400))
	mlab.clf()

	r_plot = environment.cb_radius

	rs = []
	for orbit in orbits:
		rs.append(orbit.rs)

	if show_body:
	# Plot central body
		plot_body(environment)


	# plot the x,y,z vectors
	l = r_plot*2.0
	x, y, z = [[0,0,0],[0,0,0],[0,0,0]]
	u, v, w = [[l,0,0],[0,l,0],[0,0,l]]
	mlab.quiver3d(x,y,z,u,v,w,color=(1,0,0))

	# Check for custom axes limits
	max_val = np.max(np.abs(rs))

	# Plot trajectory and starting point
	for r in rs:
		# print(r)
		mlab.plot3d(r[:,0],r[:,1],r[:,2],color = (random.uniform(0.2, 1),
					random.uniform(0.2, 1),random.uniform(0.2, 1)),
										tube_radius=None)
	
	# Check for custom axes limits
	max_x = np.max([r[:,0] for r in rs])
	min_x = np.min([r[:,0] for r in rs])
	max_y = np.max([r[:,1] for r in rs])
	min_y = np.min([r[:,1] for r in rs])
	max_z = np.max([r[:,2] for r in rs])
	min_z = np.min([r[:,2] for r in rs])

	if np.abs(min_z) < r_plot:
		min_z = -r_plot
	if max_z < r_plot:
		max_z = r_plot

	extent=[min_x,max_x,min_y,max_y,min_z,max_z]
	axes = mlab.axes(extent=[min_x,max_x,min_y,max_y,min_z,max_z]
	 				,xlabel = 'X (km)', ylabel = 'Y (km)', zlabel='Z (km)'
	 				,nb_labels=5)

	axes.label_text_property.font_family = 'courier'
	axes.label_text_property.font_size = 6
	axes.title_text_property.font_family = 'courier'
	axes.title_text_property.font_size = 8

	if show_plot:
		mlab.show()
	if save_plot:
		mlab.savefig(title+'.png',dpi=300)

def osv_to_koe(osv, mu):
	'''
	input an orbital state vector
	returns a corresponding vector of keplerian orbital elements
	'''
	r_vec = osv[:3]
	v_vec = osv[3:]
	h_bar = np.cross(r_vec, v_vec)
	h = np.linalg.norm(h_bar)
	r = np.linalg.norm(r_vec)
	v = np.linalg.norm(v_vec)
	E=0.5 * v**2 - mu/r
	a = -mu/(2*E)
	e = np.sqrt(1 - h**2 / (a*mu))
	i = np.arccos(h_bar[2]/h)
	if i == 0:
		lan = 0
		lat = 0
	else:
		lan = np.arctan2(h_bar[0], -h_bar[1])
		lat = np.arctan2(np.divide(r_vec[2],(np.sin(i))),(r_vec[0]*np.cos(lan) + r_vec[1]*np.sin(lan)))

	p = a*(1 - e**2)
	ta = np.arctan2(np.sqrt(p/mu) * np.dot(r_vec,v_vec), p-r)
	ap = lat - ta
	koe = a, e, i, lan, ap, ta
	return koe

def koe_to_osv(koe, mu, deg=True):
	'''
	input a vector of keplerian orbital elements
	returns the corresponding orbital state vector
	'''
	a, e, i, lan, ap, ta = koe

	if deg:
		np.radians(i)
		np.radians(lan)
		np.radians(ap)
		np.radians(ta)
	EA = 2*np.arctan(np.tan(ta/2) / np.sqrt(np.divide((1+e),(1-e))))
	r = a*(1 - e*np.cos(EA))
	h = np.sqrt(mu*a * (1 - e**2))


	x = r*(np.cos(lan)*np.cos(ap+ta) - np.sin(lan)*np.sin(ap+ta)*np.cos(i))
	y = r*(np.sin(lan)*np.cos(ap+ta) + np.cos(lan)*np.sin(ap+ta)*np.cos(i))
	z = r*(np.sin(i)*np.sin(ap+ta))

	p = a*(1-e**2)

	u = (x*h*e/(r*p))*np.sin(ta) - (h/r)*(np.cos(lan)*np.sin(ap+ta) + \
	np.sin(lan)*np.cos(ap+ta)*np.cos(i))
	v = (y*h*e/(r*p))*np.sin(ta) - (h/r)*(np.sin(lan)*np.sin(ap+ta) - \
	np.cos(lan)*np.cos(ap+ta)*np.cos(i))
	w = (z*h*e/(r*p))*np.sin(ta) + (h/r)*(np.cos(ap+ta)*np.sin(i))


	osv = x, y, z, u, v, w
	return osv

def tle_to_koe(tle_filename,mu):
	'''
	input a file containing a single set wo-line elements
	returns a the keplerian orbital elements
	'''

	# read in TLE file
	with open(tle_filename, 'r') as f:
		lines = f.readlines()

	# separate into three lines
	line0 = lines[0].strip() # satellite name
	line1 = lines[1].strip().split()
	line2 = lines[2].strip().split()

	# epoch (year and day)
	epoch = line1[3]
	year, month, day, hour = calc_epoch(epoch)

	# process koe

	i = float(line2[2])
	lan = float(line2[3])
	e_string = line2[4]
	e = float('0.' + e_string)
	ap = float(line2[5])
	Me = float(line2[6]) 
	mean_motion = float(line2[7])
	T = 1/mean_motion * 24*3600
	a = (T**2*mu/4.0/np.pi**2)**(1/3.0)

	E = kp.solve(Me,e)
	ta = 2*np.arctan(np.sqrt(np.divide(1+e, 1-e) * np.tan(E/2)))

	return a, e, i, lan, ap, ta, [year, month, day, hour]

def tle_to_osv(tle_filname):
	'''
	input a file containing a single set of two-line elements
	returns an orbital state vector
	'''
	return koe_to_osv(tle2koe(tle_filename), deg=True)

def calc_epoch(epoch):
	year = int('20' + epoch[:2])
	epoch = epoch[2:].split('.')
	day_of_year = int(epoch[0])-1
	hour = float('0.' + epoch[1]) * 24.0
	date = datetime.date(year,1,1) + datetime.timedelta(day_of_year)

	month = float(date.month)
	day = float(date.day)
	
	return year, month, day, hour

def tles_to_koes(tle_filename,mu):
	'''
	input a file containing sets of two-line elements
	returns a list of sets of keplerian orbital elements
	'''
	tles_lst = get_tles(tle_filename)

	koes = []
	for tle_str in tles_lst:
		lines = tle_str

		# separate into three lines
		line0 = lines[0].strip() # satellite name
		line1 = lines[1].strip().split()
		line2 = lines[2].strip().split()

		# epoch (year and day)
		epoch = line1[3]
		year, month, day, hour = calc_epoch(epoch)

		# Process koe

		i = float(line2[2])
		lan = float(line2[3])
		e_string = line2[4]
		e = float('0.' + e_string)
		ap = float(line2[5])
		Me = float(line2[6]) 
		mean_motion = float(line2[7])
		T = 1/mean_motion * 24*3600
		a = (T**2*mu/4.0/np.pi**2)**(1/3.0)

		E = np.degrees(kp.solve(np.radians(Me),e))
		ta = 2*np.degrees(np.arctan(np.sqrt(np.divide(1+e, 1-e)) * np.tan(E/2)))

		koes.append([a, e, i, lan, ap, ta])
	return koes		
		
def get_tles(tle_filename):
	# read in TLE file
	with open(tle_filename, 'r') as f:
		lines = f.readlines()
	tles_lst = chunks(lines,3)
	return tles_lst

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def plot_body(environment):
	# Display continents outline, using the VTK Builtin surface 'Earth'
	from mayavi.sources.builtin_surface import BuiltinSurface
	continents_src = BuiltinSurface(source='earth', name='Continents')
	# The on_ratio of the Earth source controls the level of detail of the
	# continents outline.
	continents_src.data_source.on_ratio = 2
	continents_src.data_source.radius = environment.cb_radius
	continents = mlab.pipeline.surface(continents_src, color=(0, 0, 0))

	# Display a sphere, for the surface of the Earth

	sphere = mlab.points3d(0, 0, 0, environment.cb_radius, 
									scale_mode='scalar',
	                                scale_factor=2,
	                                color=(0.67, 0.77, 0.93),
	                                resolution=50,
	                                opacity=1,
	                                name='Earth')

	sphere.actor.property.specular = 0.45
	sphere.actor.property.specular_power = 5

	# Plot the equator and the tropiques
	theta = np.linspace(0, 2 * np.pi, 100)
	for angle in (- np.pi / 6, 0, np.pi / 6):
	    x = environment.cb_radius*np.cos(theta) * np.cos(angle)
	    y = environment.cb_radius*np.sin(theta) * np.cos(angle)
	    z = environment.cb_radius*np.ones_like(theta) * np.sin(angle)

	    mlab.plot3d(x, y, z, color=(0,0,0),
	                        opacity=1, tube_radius=None)



# -*- coding: utf-8 -*-

"""Orbit Class"""

import numpy as np
from scipy.integrate import ode
import util as ut
import pyatmos as atm
import mayavi as mya

class Orbit:
    """
    A class to repersent the orbit of the spacecraft.

    ...

    Attributes
    ---------
    spacecraft : Spacecraft
        the spacecraft that will have its orbit calculated
    environment : Environment
        the environment settings for the simulation
    tspan : int
        span of time that the simulation runs for
    dt : int
        time step for the simulation
    """

    def __init__(
            self,
            spacecraft,
            environment,
            tspan,
            dt
    ):
        self.spacecraft = spacecraft
        self.environment = environment
        self.tspan = tspan
        self.dt = dt

        # convert keplerian orbital elements to orbital state vector if necessary
        if self.spacecraft.state_format == 'koe':
            self.spacecraft.set_state0(ut.koe_to_osv(
                self.spacecraft.state0,self.environment.cb_mu))

        # calculate number of steps required
        self.n_steps = int(np.ceil(self.tspan/self.dt))

        # allocate memory for time and state arrays
        self.ts = np.zeros((self.n_steps,1))
        self.ys = np.zeros((self.n_steps,6))

        # initial conditions
        self.ys[0] = self.spacecraft.state0
        self.step = 1

        # set up solver
        self.solver = ode(self.diff_y)
        self.solver.set_integrator('lsoda')
        self.solver.set_initial_value(self.ys[0],self.ts[0])

        self.solve_orbit()


    def solve_orbit(self):
        # propagate the orbit
        while self.solver.successful() and self.step < self.n_steps:
            self.solver.integrate(self.solver.t+self.dt)
            self.ts[self.step] = self.solver.t
            self.ys[self.step] = self.solver.y
            self.step+=1

        self.rs = self.ys[:,:3]
        self.vs = self.ys[:,3:]


    def diff_y(self,t,y):

        r = y[:3]
        v = y[3:]
        r_norm = np.linalg.norm(r)

        # two body acceleration
        a = -r*self.environment.cb_mu / r_norm**3

        # constant thrust
        if "thrust" in self.environment.perturbations:
            v_dir = v/np.linalg.norm(v)
            a_thrust = v_dir * self.spacecraft.engine.thrust/self.spacecraft.mass_total

            a+= a_thrust

        # aerodynamic drag
        if 'aero' in self.environment.perturbations:
            #calculate air density
            z = r_norm - self.environment.cb_radius  # km

            rho = atm.coesa76(z)[0][0]*1e9  # kg / km^3

            #calculate motion of s/c wrt a rotating atmosphere
            v_rel = v - np.cross(self.environment.cb_atm_rot,r)

            drag = 0.5 * rho * v_rel*np.linalg.norm(v_rel) * self.spacecraft.drag_coeff\
            * self.spacecraft.area / self.spacecraft.mass_total

            a += drag

        # J2 perturbation
        if 'J2' in self.environment.perturbations:

            z2 = r[2]**2
            r2 = r_norm**2
            tx = r[0] / r_norm*(5*z2/r2 -1)
            ty = r[1] / r_norm*(5*z2/r2 -1)
            tz = r[2] / r_norm*(5*z2/r2 -3)

            a_j2 = 1.5*self.environment.cb_J2*self.environment.cb_mu\
            * self.environment.cb_radius**2 / r_norm**4 * np.array([tx, ty, tz])

            a += a_j2

        return np.concatenate([v,a])




# -*- coding: utf-8 -*-

"""Environment Class"""

__author__ = "Alexander William Minchin"

import json
from scipy import constants as const

CONFIG_FILE = '../simulator_values/celestial_bodies.json'

class Environment:
    """
    A class to repersent the environment setup

    ...

    Attributes
    ---------
	celestial_body : String
    	name of celestrial body that has been defined in
    	celestial_bodies.json
    perturbations : list
    	list of perturbations to be included in the simulation
    	['thrust', 'aero', 'J2']
    """

    def __init__(
            self,
            celestial_body,
            perturbations=[]
    ):

	    """
	        Parameters
	        ----------
	        celestial_body : String
	        	name of celestrial body that has been defined in
	        	celestial_bodies.json
	        perturbations : list
	        	list of perturbations to be included in the simulation
	        	['thrust', 'aero', 'J2']
	    """

	    self.celestial_body = celestial_body
	    self.perturbations = perturbations

	    self.get_celestial_body()

    def get_celestial_body(self):
    	"""read in data about celestial body and unpack"""

    	#read in json file
    	with open(CONFIG_FILE) as f:
    		data = json.load(f)

    	celestial_body_dict = data[self.celestial_body]

    	# unpack json file
    	self.cb_mass = celestial_body_dict['mass_kg']
    	self.cb_radius = celestial_body_dict['radius_m']
    	self.cb_J2 = celestial_body_dict['J2']
    	self.cb_atm_rot = celestial_body_dict['atm_rot_vec_rad_p_s']

    	# calculate parameters
    	self.cb_mu = const.G * self.cb_mass # m^3 kg^-1 s^-2





import environment, engine, spacecraft, orbit 
import util as t
import numpy as np

if __name__ == '__main__':
	
	#initialise environment object
	env = environment.Environment("Earth")

	# initialise spacecraft
	eng1 = engine.Engine(100)
	sc1 = spacecraft.Spacecraft(eng1)
	sc1.set_state0([50e6,0.1,0,0,0,0], 'koe')

	eng2 = engine.Engine(200)
	sc2 = spacecraft.Spacecraft(eng2)
	sc2.set_state0([40e6,0.3,np.radians(25),0,0,0], 'koe')

	# define simulation parameters
	tspan = 3*24*60*60 # 1 day in seconds
	dt = 600 # 10 minutes

	# calcualte orbits
	orb1= orbit.Orbit(sc1,env,tspan,dt)
	orb2 = orbit.Orbit(sc2,env,tspan,dt)

	# plot orbits
	t.plot_n_orbits([orb1,orb2],env)


# -*- coding: utf-8 -*-

"""Engine Class"""

__author__ = "Alexander William Minchin"

from scipy import constants as const


class Engine:
    """
    A class to repersent the engine of the spacecraft.

    ...

    Attributes
    ---------
    mass : float
        dry mass of engine (kg)
    mass_p : float
        mass of propellant (kg)
    isp : float
        the specific impulse of the engine (N*s/kg)
    thrust : float
        thrust produced by the enginer (N)
    mass_p_dot : float
        time derivative of proellant mass (kg/s)
    burnout_time : float
        time till all propellant is burnt (s)
    """

    def __init__(
            self,
            mass,
            mass_p=0,
            isp=0,
            thrust=0
    ):
        """
        Parameters
        ----------
        mass : float
            dry mass of engine (kg)
        mass_p : float
            mass of propellant (kg)
        isp : float
            the specific impulse of the engine (N*s/kg)
        thrust : float
            thrust produced by the enginer (N)

        """
        # define the engine attributes in SI units
        self.mass = mass
        self.mass_p = mass_p
        self.isp = isp
        self.thrust = thrust

        # attributes to be calculated
        if self.thrust > 0:
            self.mass_dot = self.calculate_mass_dot()
            self.burnout_time = self.calculate_burnout_time()

    def calculate_mass_dot(self):
        """ calculates and returns the time derivative of proellant mass (kg/s)"""

        #calculate m dot
        self.mass_p_dot = self.thrust / (const.g * self.isp)

        return self.mass_p_dot

    def calculate_burnout_time(self):
        """ calculates the time till all propellant is burnt (s)"""

        #calculate burnout time
        self.burnout_time = self.mass_p / self.mass_p_dot

        return self.burnout_time

    def info(self):
        """ prints information summary of the engine"""

        #print engine information
        print("\nEngine Information")
        print("Engine Mass:", self.mass, "kg")
        print("Engine Isp:", self.isp, "N*s/kg")
        print("Engine Thrust:", self.thrust, "N")
        print("Propellant Mass:", self.mass_p, "kg")
        print("Engine Burnout Time", self.burnout_time, "s")



# -*- coding: utf-8 -*-

# Learn more: https://github.com/kennethreitz/setup.py

from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()
setup(
    name='orbitsim',
    version='0.0.1b',
    description='3D orbit simulation visualisation tool',
    long_description=readme,
    author='Alexander Minchin',
    author_email='alexander.w.minchin@gmail.com',
    url='https://github.com/alexander-minchin/orbitsim',
    license=license,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={'':'src'},
    packages=find_packages(where="src",exclude=('tests', 'docs')),
    install_requires=requirements,
    python_requires=">=3.6"
)


# -*- coding: utf-8 -*-

"""Spacecraft Class"""

__author__ = "Alexander William Minchin"


class Spacecraft:
    """
    A class to repersent the spacecraft in orbit.

    ...

    Attributes
    ---------
    mass : float
        structural mass of spacecraft (kg)
    engine : Engine
        the engine that is used by the spacecraft
    drag_coeff : float
        the coefficient of drag for the spacecraft
    area : float
        the cross sectional area of the spacecraft (m^2)
    mass_total : float
        total mass of the spacecraft including engine and propellant (kg)
    """

    def __init__(
            self,
            engine,
            mass=3000.0,
            drag_coeff=2.2,
            area=6.0
    ):
        """
        Parameters
        ----------
        mass : float
            mass of spacecraft without fuel loaded (kg)
        engine : Engine
            the engine that is used by the spacecraft
        drag_coeff : float
            the coefficient of drag for the spacecraft (-)
        area : float
            the cross sectional area of the spacecraft (m^2)
        state0 : array
            state vector containing the initial 3D position and velocities
            [x,y,z,u,v,w]       [(m),(m),(m),(m/s),(m/s),(m/s)]
        state_format : string
            string defining the format of the input state vector 'osv' or 'koe'
            koe format = [a,e,i,lan,aop,ta]
        """
        # define the spacecraft attributes in SI units
        self.mass = mass
        self.engine = engine
        self.drag_coeff = drag_coeff
        self.area = area

        # attributes to be set or caclulated
        self.state0 = None
        self.state_format = None

        #attributed to be calculated
        self.mass_total = self.calculate_total_mass()

    def calculate_total_mass(self):
        """ calculates and returns the total mass of the spacecraft
        including engine mass and propellant mass"""

        #calcualte mass by summing spacecraft, engine and propellant masses
        self.mass_total = self.mass + self.engine.mass + self.engine.mass_p

        return self.mass_total

    def set_state0(self, state0, state_format='osv'):
        """ sets the state vector in orbital state vector format"""

        self.state0 = state0
        self.state_format = state_format

    def info(self):
        """ prints information summary of the engine"""

        #print engine information
        print("\nSpacecraft Information")
        print("Spacecraft Mass:", self.mass, "kg")
        print("Spacecraft Cd:", self.drag_coeff)
        print("Spacecraft Cross-sectional Area:", self.area, "m^2")
        print("Spacecraft Total Mass:", self.mass_total, "kg")
        print("Spacecraft Orbital State Vector", self.state0)
